
/* xmount -- device mounter for X
   Copyright (C) 1996 Kidong Lee, Thomas Uhl

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>

#include <unistd.h>
#include <mntent.h>
#include <sys/mount.h>
#include <linux/fs.h>

#define VERSION "1.0"

#define MOUNTED_LOCK "/etc/mtab~"
#define CONFIG_FILE "xmounttab"
#define XMOUNTPATH "XMOUNTPATH"

#define BOOL int
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define STRLEN 80
#define LINELEN 255
#define MAXITEMS 16

#define streq(s, t)     (strcmp ((s), (t)) == 0)

/* types */
typedef struct
  {
    char name[STRLEN];
    char dev[STRLEN];
    char dir[STRLEN];
    char type[STRLEN];
    char opt[STRLEN];
    BOOL mounted;
    Widget Name, Label, Button;
  }
MITEM;

typedef struct
  {
    const char *opt;		/* option name */
    int inv;			/* true if flag value should be inverted */
    int mask;			/* flag mask value */
  }
OPTMAP;


/* global variables */
Widget status_box, msg_box, msg_first, msg_second;
int devices;			/* number of devices */
MITEM mitems[MAXITEMS];		/* device list */

const OPTMAP opt_map[] =
{
  {"defaults", 0, 0},		/* default options */
  {"ro", 0, MS_RDONLY},		/* read-only */
  {"rw", 1, MS_RDONLY},		/* read-write */
  {"exec", 1, MS_NOEXEC},	/* permit execution of binaries */
  {"noexec", 0, MS_NOEXEC},	/* don't execute binaries */
  {"suid", 1, MS_NOSUID},	/* honor suid executables */
  {"nosuid", 0, MS_NOSUID},	/* don't honor suid executables */
  {"dev", 1, MS_NODEV},		/* interpret device files  */
  {"nodev", 0, MS_NODEV},	/* don't interpret devices */
  {"sync", 0, MS_SYNCHRONOUS},	/* synchronous I/O */
  {"async", 1, MS_SYNCHRONOUS},	/* asynchronous I/O */
#ifdef MS_NOSUB
  {"sub", 1, MS_NOSUB},		/* allow submounts */
  {"nosub", 0, MS_NOSUB},	/* don't allow submounts */
#endif
  /* add new options here */
  {NULL, 0, 0}
};


void
parse_opt (char *opt, int *mask, char *extra_opts)
{
  const OPTMAP *om;

  for (om = opt_map; om->opt != NULL; ++om)
    if (streq (opt, om->opt))
      {
	if (om->inv)
	  *mask &= ~om->mask;
	else
	  *mask |= om->mask;
	return;
      }

  if (*extra_opts)
    strcat (extra_opts, ",");
  strcat (extra_opts, opt);
}


void
parse_opts (char *opts, int *flags, char *extra_opts)
{
  char *opt, *lopts;

  lopts = strdup (opts);
  assert (lopts);

  *extra_opts = '\0';

  if (!streq (lopts, "defaults"))
    {
      for (opt = strtok (lopts, ",");
	   opt != NULL;
	   opt = strtok (NULL, ","))
	parse_opt (opt, flags, extra_opts);
    }

  free (lopts);
}

/*ARGSUSED */
void 
PrintErrorMsg (char *msg)
{
  XtVaSetValues (msg_first, XtNlabel, "", NULL);
  XtVaSetValues (msg_second, XtNlabel, msg, NULL);
}


BOOL
do_mount (char *dev, char *mdir, char *opt, char *type, Widget item)
{
  int flags = 0;
  char extra_opts[STRLEN];
  struct mntent mnt;
  FILE *F_mtab;
  char msg[STRLEN];

  sprintf (msg, "Mounted Filesystem Type: %s", type);
  XtVaSetValues (msg_first, XtNlabel, msg, NULL);

  parse_opts (opt, &flags, extra_opts);

#ifdef DEBUG
  printf ("mount (<%s>, <%s>, <%s>, %d, <%s>)\n",
	  dev, mdir, type, flags, extra_opts);
#endif

#ifndef DONT_MOUNT
  if (mount (dev, mdir, type, flags | MS_MGC_VAL, extra_opts) < 0)
    {
      sprintf (msg, "Mount: %s", strerror (errno));
      PrintErrorMsg (msg);
      return FALSE;
    }

  if ((F_mtab = setmntent (MOUNTED, "a+")) == NULL)
    {
      sprintf (msg, "Can't open %s: %s", MOUNTED, strerror (errno));
      PrintErrorMsg (msg);
      /* simulate successful operation */
      return TRUE;
    }

  mnt.mnt_fsname = dev;
  mnt.mnt_dir = mdir;
  mnt.mnt_type = type;
  mnt.mnt_opts = opt;
  mnt.mnt_freq = 0;
  mnt.mnt_passno = 0;
  addmntent (F_mtab, &mnt);
  endmntent (F_mtab);
#endif

  sprintf (msg, "%s mounted", dev);
  XtVaSetValues (msg_second, XtNlabel, msg, NULL);
  XtVaSetValues (item, XtNlabel, type, NULL);

  return TRUE;
}


BOOL
do_umount (char *dev, Widget item)
{
  FILE *F_mtab, *F_lock;
  struct mntent *mnt;
  char msg[80];

  sync ();

#ifdef DEBUG
  printf ("umount (<%s>)\n", dev);
#endif

#ifndef DONT_MOUNT
  if (umount (dev) < 0)
    {
      sprintf (msg, "Unmount: %s", strerror (errno));
      PrintErrorMsg (msg);
      return FALSE;
    }

  if ((F_mtab = setmntent (MOUNTED, "r")) == NULL)
    {
      sprintf (msg, "Can't open %s: %s", MOUNTED, strerror (errno));
      PrintErrorMsg (msg);
      /* simulate successful operation */
      return TRUE;
    }
  if ((F_lock = setmntent (MOUNTED_LOCK, "w+")) == NULL)
    {
      sprintf (msg, "Can't open %s: %s", MOUNTED_LOCK, strerror (errno));
      PrintErrorMsg (msg);
      /* simulate successful operation */
      return TRUE;
    }

  while ((mnt = getmntent (F_mtab)) != NULL)
    if (strcmp (dev, mnt->mnt_fsname))
      addmntent (F_lock, mnt);
  endmntent (F_mtab);
  endmntent (F_lock);

  if (rename (MOUNTED_LOCK, MOUNTED) < 0)
    {
      sprintf (msg, "Can't rename %s to %s: %s", MOUNTED_LOCK, MOUNTED,
	       strerror (errno));
      PrintErrorMsg (msg);
      /* simulate successful operation */
      return TRUE;
    }
#endif
  XtVaSetValues (msg_first, XtNlabel, "", NULL);
  sprintf (msg, "%s unmounted", dev);
  XtVaSetValues (msg_second, XtNlabel, msg, NULL);
  XtVaSetValues (item, XtNlabel, "unmounted", NULL);

  return TRUE;
}


void
ButtonClick (w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
  MITEM *mitem;

  mitem = client_data;

  if (!mitem->mounted)
    {
      if ((mitem->mounted = do_mount (mitem->dev,
				     mitem->dir,
				     mitem->opt,
				     mitem->type,
				     mitem->Label)) == TRUE)
	XtVaSetValues (w, XtNlabel, "umount", NULL);
    }
  else
    {
      if (!(mitem->mounted = !do_umount (mitem->dev, mitem->Label)))
	XtVaSetValues (w, XtNlabel, "mount", NULL);
    }
}


void
check_mounted_fs ()
{
  int i;
  FILE *F_mtab;
  struct mntent *mnt;
  char msg[80];

  if ((F_mtab = setmntent (MOUNTED, "r")) == NULL)
    {
      sprintf (msg, "Can't open %s: %s", MOUNTED, strerror (errno));
      PrintErrorMsg (msg);
      return ;
    }

  while ((mnt = getmntent (F_mtab)) != NULL)
    {
      for (i = 0; i < devices; i++)
	{
	  if (!strcmp (mitems[i].dev, mnt->mnt_fsname))
	    {
	      /* fs is already mounted */
	      mitems[i].mounted = TRUE;
	      XtVaSetValues (mitems[i].Label,
			     XtNlabel, mnt->mnt_type,
			     XtNx, 200, NULL);
	      XtVaSetValues (mitems[i].Button,
			     XtNlabel, "umount", NULL);
	    }
	}
    }

  endmntent (F_mtab);
}


void
CreateButtons ()
{
  int i, lin = 1, res;
  FILE *fd;
  char *cp;
  char path[LINELEN];
  char line[LINELEN];

  /* Create file path for xmounttab */
  cp = getenv (XMOUNTPATH);
  if (cp != NULL)
    {
      res = strlen (cp) - 1;
      if (cp[res] == '/')
	cp[res] = '\0';
      sprintf (path, "%s/%s", cp, CONFIG_FILE);
    }
  else
    {
      sprintf (path, "/etc/%s", CONFIG_FILE);
    }

  devices = 0;

  /* Read devices in xmounttab */
  if ((fd = fopen (path, "r")) != NULL)
    {
      while ((cp = fgets (line, LINELEN, fd)) && devices < MAXITEMS)
	{
	  while (*cp)
	    {
	      if (*cp == '#' || *cp == '\n')
		*cp = '\0';
	      cp++;
	    }

	  if (line[0])
	    {
	      res = sscanf (line, "%s %s %s %s %s", mitems[devices].name,
			    mitems[devices].dev,
			    mitems[devices].dir,
			    mitems[devices].type,
			    mitems[devices].opt);
	      if (res > 0 && res < 5)
		{
		  fprintf (stderr, "xmount: error in line %d\n", lin);
		  exit (1);
		}

#ifdef DEBUG
	      printf ("%s - %s - %s - %s - %s\n", mitems[devices].name,
		      mitems[devices].dev,
		      mitems[devices].dir,
		      mitems[devices].type,
		      mitems[devices].opt);
#endif

	      devices++;
	    }
	  lin++;
	}
      fclose (fd);
    }
  else
    {
      fprintf (stderr, "xmount: %s: %s\n", path, strerror (errno));
      exit (1);
    }

  /* Create buttons in box */
  for (i = 0; i < devices; i++)
    {
      mitems[i].mounted = FALSE;
      mitems[i].Name = XtVaCreateManagedWidget ("name",
						labelWidgetClass,
						status_box,
						XtNlabel, mitems[i].name,
						NULL);

      mitems[i].Label = XtVaCreateManagedWidget ("label",
						 labelWidgetClass,
						 status_box,
						 XtNlabel, "unmounted",
						 NULL);

      mitems[i].Button = XtVaCreateManagedWidget ("button",
						  commandWidgetClass,
						  status_box,
						  XtNlabel, "mount",
						  XtNx, 170,
						  XtNwidth, 60,
						  NULL);

      XtAddCallback (mitems[i].Button, XtNcallback, ButtonClick, &mitems[i]);

    }

  check_mounted_fs ();
}

/*ARGSUSED */
void 
Quit (w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
  exit (0);
}

void
main (int argc, char *argv[])
{
  XtAppContext app_context;
  Widget top, paned, menu, quit;


  top = XtVaAppInitialize (&app_context, "XMount", NULL, 0,
			   &argc, argv, NULL, NULL);

  paned = XtVaCreateManagedWidget ("paned", panedWidgetClass,
				   top, NULL);

  menu = XtVaCreateManagedWidget ("menu", boxWidgetClass, paned, NULL);

  quit = XtVaCreateManagedWidget ("quit", commandWidgetClass, menu, 
				  XtNlabel, "Quit", NULL);
  XtAddCallback (quit, XtNcallback, Quit, 0);

  status_box = XtVaCreateManagedWidget ("status_box", boxWidgetClass, paned, 
					XtNwidth, 300, NULL);

  CreateButtons ();

  msg_box = XtVaCreateManagedWidget ("msg_box", boxWidgetClass, paned, 
				     XtNwidth, 300, NULL);

  msg_first = XtVaCreateManagedWidget ("msg", labelWidgetClass, msg_box,
				       XtNlabel, "", 
				       NULL);
  
  msg_second = XtVaCreateManagedWidget ("msg", labelWidgetClass, 
					msg_box,
					XtNlabel, "",
					NULL);
  XtRealizeWidget (top);
  XtAppMainLoop (app_context);
}
